@@ -31,17 +31,14 @@ try:
     import hmac
 except ImportError:
     import _hmac as hmac
-import sha
 import socket
 import struct
 import threading
-import logging
 from types import StringType
 
-from ZODB.loglevels import TRACE
-
-from ZEO.zrpc.log import log, short_repr
+from ZEO.zrpc.log import log
 from ZEO.zrpc.error import DisconnectedError
+import ZEO.hash
 
 
 # Use the dictionary to make sure we get the minimum number of errno
@@ -147,8 +144,8 @@ class SizedMessageAsyncConnection(asyncore.dispatcher):
         # and thus iterator, because it contains a yield statement.
 
         def hack():
-            self.__hmac_send = hmac.HMAC(sesskey, digestmod=sha)
-            self.__hmac_recv = hmac.HMAC(sesskey, digestmod=sha)
+            self.__hmac_send = hmac.HMAC(sesskey, digestmod=ZEO.hash)
+            self.__hmac_recv = hmac.HMAC(sesskey, digestmod=ZEO.hash)
             if False:
                 yield ''
 
@@ -298,11 +295,17 @@ class SizedMessageAsyncConnection(asyncore.dispatcher):
             i += 1
             # It is very unlikely that i will be 1.
             v = "".join(output[:i])
+            # Note: "output" usually contains the output not yet sent
+            #  The "del" below breaks this invariant temporarily.
+            #  We must ensure its satisfaction again when we leave the loop
             del output[:i]
 
             try:
                 n = self.send(v)
             except socket.error, err:
+                # Fix for https://bugs.launchpad.net/zodb/+bug/182833
+                #  ensure the above mentioned "output" invariant
+                output.insert(0, v)
                 if err[0] in expected_socket_write_errors:
                     break # we couldn't write anything
                 raise
