@@ -23,19 +23,18 @@ from ZODB.config import databaseFromString
 from ZODB.utils import p64, u64
 from ZODB.tests.warnhook import WarningsHook
 from zope.interface.verify import verifyObject
+import ZODB.tests.util
 
-class ConnectionDotAdd(unittest.TestCase):
+class ConnectionDotAdd(ZODB.tests.util.TestCase):
 
     def setUp(self):
+        ZODB.tests.util.TestCase.setUp(self)
         from ZODB.Connection import Connection
         self.db = StubDatabase()
         self.datamgr = Connection(self.db)
         self.datamgr.open()
         self.transaction = StubTransaction()
 
-    def tearDown(self):
-        transaction.abort()
-
     def check_add(self):
         from ZODB.POSException import InvalidObjectReference
         obj = StubObject()
@@ -97,7 +96,7 @@ class ConnectionDotAdd(unittest.TestCase):
         self.assert_(obj._p_oid is None)
         self.assert_(obj._p_jar is None)
         self.assertRaises(KeyError, self.datamgr.get, oid)
-        self.assertEquals(self.db._storage._stored, [oid])
+        self.assertEquals(self.db.storage._stored, [oid])
 
     def checkCommit(self):
         obj = StubObject()
@@ -112,8 +111,8 @@ class ConnectionDotAdd(unittest.TestCase):
         # This next assert_ is covered by an assert in tpc_finish.
         ##self.assert_(not self.datamgr._added)
 
-        self.assertEquals(self.db._storage._stored, [oid])
-        self.assertEquals(self.db._storage._finished, [oid])
+        self.assertEquals(self.db.storage._stored, [oid])
+        self.assertEquals(self.db.storage._finished, [oid])
 
     def checkModifyOnGetstate(self):
         member = StubObject()
@@ -124,7 +123,7 @@ class ConnectionDotAdd(unittest.TestCase):
         self.datamgr.tpc_begin(self.transaction)
         self.datamgr.commit(self.transaction)
         self.datamgr.tpc_finish(self.transaction)
-        storage = self.db._storage
+        storage = self.db.storage
         self.assert_(obj._p_oid in storage._stored, "object was not stored")
         self.assert_(subobj._p_oid in storage._stored,
                 "subobject was not stored")
@@ -153,9 +152,6 @@ class UserMethodTests(unittest.TestCase):
     # add isn't tested here, because there are a bunch of traditional
     # unit tests for it.
 
-    # The version tests would require a storage that supports versions
-    # which is a bit more work.
-
     def test_root(self):
         r"""doctest of root() method
 
@@ -220,12 +216,12 @@ class UserMethodTests(unittest.TestCase):
         >>> obj._p_state
         0
 
-        A request for an object that doesn't exist will raise a KeyError.
+        A request for an object that doesn't exist will raise a POSKeyError.
 
         >>> cn.get(p64(1))
         Traceback (most recent call last):
           ...
-        KeyError: '\x00\x00\x00\x00\x00\x00\x00\x01'
+        POSKeyError: 0x01
         """
 
     def test_close(self):
@@ -356,7 +352,7 @@ class UserMethodTests(unittest.TestCase):
 
         An expedient way to create a read-only storage:
 
-        >>> db._storage._is_read_only = True
+        >>> db.storage.isReadOnly = lambda: True
         >>> cn = db.open()
         >>> cn.isReadOnly()
         True
@@ -523,6 +519,268 @@ def test_invalidateCache():
         >>> db.close()
     """
 
+def connection_root_convenience():
+    """Connection root attributes can now be used as objects with attributes
+
+    >>> db = ZODB.tests.util.DB()
+    >>> conn = db.open()
+    >>> conn.root.x
+    Traceback (most recent call last):
+    ...
+    AttributeError: x
+
+    >>> del conn.root.x
+    Traceback (most recent call last):
+    ...
+    AttributeError: x
+
+    >>> conn.root()['x'] = 1
+    >>> conn.root.x
+    1
+    >>> conn.root.y = 2
+    >>> sorted(conn.root().items())
+    [('x', 1), ('y', 2)]
+
+    >>> conn.root
+    <root: x y>
+
+    >>> del conn.root.x
+    >>> sorted(conn.root().items())
+    [('y', 2)]
+
+    >>> conn.root.rather_long_name = 1
+    >>> conn.root.rather_long_name2 = 1
+    >>> conn.root.rather_long_name4 = 1
+    >>> conn.root.rather_long_name5 = 1
+    >>> conn.root
+    <root: rather_long_name rather_long_name2 rather_long_name4 ...>
+    """
+
+class C_invalidations_of_new_objects_work_after_savepoint(Persistent):
+    def __init__(self):
+        self.settings = 1
+
+    def _p_invalidate(self):
+        print 'INVALIDATE', self.settings
+        Persistent._p_invalidate(self)
+        print self.settings   # POSKeyError here
+
+def abort_of_savepoint_creating_new_objects_w_exotic_invalidate_doesnt_break():
+    r"""
+    Before, the following would fail with a POSKeyError, which was
+    somewhat surprizing, in a very edgy sort of way. :)
+
+    Really, when an object add is aborted, the object should be "removed" from
+    the db and its invalidatuon method shouldm't even be called:
+
+    >>> conn = ZODB.connection('data.fs')
+    >>> conn.root.x = x = C_invalidations_of_new_objects_work_after_savepoint()
+    >>> _ = transaction.savepoint()
+    >>> x._p_oid
+    '\x00\x00\x00\x00\x00\x00\x00\x01'
+
+    >>> x._p_jar is conn
+    True
+
+    >>> transaction.abort()
+
+After the abort, the oid and jar are None:
+
+    >>> x._p_oid
+    >>> x._p_jar
+
+Cleanup:
+
+    >>> conn.close()
+
+    """
+
+class Clp9460655(Persistent):
+    def __init__(self, word, id):
+        super(Clp9460655, self).__init__()
+	self.id = id
+        self._word = word
+
+def lp9460655():
+    r"""
+    >>> conn = ZODB.connection('data.fs')
+    >>> root = conn.root()
+    >>> Word = Clp9460655
+
+    >>> from BTrees.OOBTree import OOBTree
+    >>> data = root['data'] = OOBTree()
+
+    >>> commonWords = []
+    >>> count = "0"
+    >>> for x in ('hello', 'world', 'how', 'are', 'you'):
+    ...         commonWords.append(Word(x, count))
+    ...         count = str(int(count) + 1)
+
+    >>> sv = transaction.savepoint()
+    >>> for word in commonWords:
+    ...         sv2 = transaction.savepoint()
+    ...         data[word.id] = word
+
+    >>> sv.rollback()
+    >>> print commonWords[1].id  # raises POSKeyError
+    1
+
+    Cleanup:
+
+    >>> transaction.abort()
+    >>> conn.close()
+
+    """
+
+def lp615758_transaction_abort_Incomplete_cleanup_for_new_objects():
+    r"""
+
+As the following"DocTest" demonstrates, "abort" forgets to
+reset "_p_changed" for new (i.e. "added") objects.
+
+>>> class P(Persistent): pass
+...
+>>> c = ZODB.connection('t.fs')
+>>> obj = P()
+>>> c.add(obj)
+>>> obj.x = 1
+>>> obj._p_changed
+True
+>>> transaction.abort()
+>>> obj._p_changed
+False
+
+>>> c.close()
+    """
+
+
+class _PlayPersistent(Persistent):
+    def setValueWithSize(self, size=0): self.value = size*' '
+    __init__ = setValueWithSize
+
+class EstimatedSizeTests(ZODB.tests.util.TestCase):
+    """check that size estimations are handled correctly."""
+
+    def setUp(self):
+        ZODB.tests.util.TestCase.setUp(self)
+        self.db = db = databaseFromString("<zodb>\n<mappingstorage/>\n</zodb>")
+        self.conn = c = db.open()
+        self.obj = obj = _PlayPersistent()
+        c.root()['obj'] = obj
+        transaction.commit()
+
+    def test_size_set_on_write_commit(self):
+        obj, cache = self.obj, self.conn._cache
+        # we have just written "obj". Its size should not be zero
+        size, cache_size = obj._p_estimated_size, cache.total_estimated_size
+        self.assert_(size > 0)
+        self.assert_(cache_size > size)
+        # increase the size, write again and check that the size changed
+        obj.setValueWithSize(1000)
+        transaction.commit()
+        new_size = obj._p_estimated_size
+        self.assert_(new_size > size)
+        self.assertEqual(cache.total_estimated_size,
+                         cache_size + new_size - size)
+
+    def test_size_set_on_write_savepoint(self):
+        obj, cache = self.obj, self.conn._cache
+        # we have just written "obj". Its size should not be zero
+        size, cache_size = obj._p_estimated_size, cache.total_estimated_size
+        # increase the size, write again and check that the size changed
+        obj.setValueWithSize(1000)
+        transaction.savepoint()
+        new_size = obj._p_estimated_size
+        self.assert_(new_size > size)
+        self.assertEqual(cache.total_estimated_size,
+                         cache_size + new_size - size)
+
+    def test_size_set_on_load(self):
+        c = self.db.open() # new connection
+        obj = c.root()['obj']
+        # the object is still a ghost and '_p_estimated_size' not yet set
+        # access to unghost
+        cache = c._cache
+        cache_size = cache.total_estimated_size
+        obj.value
+        size = obj._p_estimated_size
+        self.assert_(size > 0)
+        self.assertEqual(cache.total_estimated_size, cache_size + size)
+        # we test here as well that the deactivation works reduced the cache
+        # size
+        obj._p_deactivate()
+        self.assertEqual(cache.total_estimated_size, cache_size)
+
+    def test_configuration(self):
+        # verify defaults ....
+        expected = 0
+        # ... on db
+        db = self.db
+        self.assertEqual(db.getCacheSizeBytes(), expected)
+        self.assertEqual(db.getHistoricalCacheSizeBytes(), expected)
+        # ... on connection
+        conn = self.conn
+        self.assertEqual(conn._cache.cache_size_bytes, expected)
+        # verify explicit setting ...
+        expected = 10000
+        # ... on db
+        db = databaseFromString("<zodb>\n"
+                                "  cache-size-bytes %d\n"
+                                "  historical-cache-size-bytes %d\n"
+                                "  <mappingstorage />\n"
+                                "</zodb>"
+                                % (expected, expected+1)
+                                )
+        self.assertEqual(db.getCacheSizeBytes(), expected)
+        self.assertEqual(db.getHistoricalCacheSizeBytes(), expected+1)
+        # ... on connectionB
+        conn = db.open()
+        self.assertEqual(conn._cache.cache_size_bytes, expected)
+        # test huge (larger than 4 byte) size limit
+        db = databaseFromString("<zodb>\n"
+                                "  cache-size-bytes 8GB\n"
+                                "  <mappingstorage />\n"
+                                "</zodb>"
+                                )
+        self.assertEqual(db.getCacheSizeBytes(), 0x1L << 33)
+
+
+    def test_cache_garbage_collection(self):
+        db = self.db
+        # activate size based cache garbage collection
+        db.setCacheSizeBytes(1)
+        conn = self.conn
+        cache = conn._cache
+        # verify the change worked as expected
+        self.assertEqual(cache.cache_size_bytes, 1)
+        # verify our entrance assumption is fullfilled
+        self.assert_(cache.total_estimated_size > 1)
+        conn.cacheGC()
+        self.assert_(cache.total_estimated_size <= 1)
+        # sanity check
+        self.assert_(cache.total_estimated_size >= 0)
+
+    def test_cache_garbage_collection_shrinking_object(self):
+        db = self.db
+        # activate size based cache garbage collection
+        db.setCacheSizeBytes(1000)
+        obj, conn, cache = self.obj, self.conn, self.conn._cache
+        # verify the change worked as expected
+        self.assertEqual(cache.cache_size_bytes, 1000)
+        # verify our entrance assumption is fullfilled
+        self.assert_(cache.total_estimated_size > 1)
+        # give the objects some size
+        obj.setValueWithSize(500)
+        transaction.savepoint()
+        self.assert_(cache.total_estimated_size > 500)
+        # make the object smaller
+        obj.setValueWithSize(100)
+        transaction.savepoint()
+        # make sure there was no overflow
+        self.assert_(cache.total_estimated_size != 0)
+        # the size is not larger than the allowed maximum
+        self.assert_(cache.total_estimated_size <= 1000)
+
 # ---- stubs
 
 class StubObject(Persistent):
@@ -555,7 +813,6 @@ class StubStorage:
 
     Only one concurrent transaction is supported.
     Voting is not supported.
-    Versions are not supported.
 
     Inspect self._stored and self._finished to see how the storage has been
     used during a unit test. Whenever an object is stored in the store()
@@ -615,7 +872,7 @@ class StubStorage:
         self._transdata.clear()
         self._transstored = []
 
-    def load(self, oid, version):
+    def load(self, oid, version=''):
         if version != '':
             raise TypeError('StubStorage does not support versions.')
         return self._data[oid]
@@ -648,7 +905,7 @@ class TestConnectionInterface(unittest.TestCase):
 class StubDatabase:
 
     def __init__(self):
-        self._storage = StubStorage()
+        self.storage = StubStorage()
 
     classFactory = None
     database_name = 'stubdatabase'
@@ -661,4 +918,5 @@ def test_suite():
     s = unittest.makeSuite(ConnectionDotAdd, 'check')
     s.addTest(doctest.DocTestSuite())
     s.addTest(unittest.makeSuite(TestConnectionInterface))
+    s.addTest(unittest.makeSuite(EstimatedSizeTests))
     return s
